@@ -48,6 +48,10 @@ class CroniterError(ValueError):
     pass
 
 
+class CroniterBadTypeRangeError(TypeError):
+    """."""
+
+
 class CroniterBadCronError(CroniterError):
     """ Syntax, unknown value, or range error within a cron expression """
     pass
@@ -708,7 +712,13 @@ class croniter(object):
                     except ValueError:
                         pass
 
-                    if t in cls.LOWMAP[i]:
+                    if t in cls.LOWMAP[i] and not (
+                        # do not support 0 as a month either for classical 5 fields cron
+                        # or 6fields second repeat form
+                        # but still let conversion happen if day field is shifted
+                        (i == 3 and len(expressions) == 5) or
+                        (i == 4 and len(expressions) == 6)
+                    ):
                         t = cls.LOWMAP[i][t]
 
                     if (
@@ -795,9 +805,15 @@ def croniter_range(start, stop, expr_format, ret_type=None, day_or=True, exclude
     """
     _croniter = _croniter or croniter
     auto_rt = datetime.datetime
-    if type(start) != type(stop):
-        raise TypeError("The start and stop must be same type.  {0} != {1}".
-                        format(type(start), type(stop)))
+    # type is used in first if branch for perfs reasons
+    if (
+        type(start) != type(stop) and not (
+            isinstance(start, type(stop)) or
+            isinstance(stop, type(start)))
+    ):
+        raise CroniterBadTypeRangeError(
+            "The start and stop must be same type.  {0} != {1}".
+            format(type(start), type(stop)))
     if isinstance(start, (float, int)):
         start, stop = (datetime.datetime.utcfromtimestamp(t) for t in (start, stop))
         auto_rt = float
