@@ -6,7 +6,8 @@ from datetime import datetime, timedelta
 from functools import partial
 from time import sleep
 import pytz
-from croniter import croniter, croniter_range, CroniterBadDateError, CroniterBadCronError, CroniterNotAlphaError
+from croniter import (croniter, CroniterBadDateError,  CroniterBadCronError,
+                      CroniterNotAlphaError, CroniterUnsupportedSyntaxError)
 from croniter.tests import base
 import dateutil.tz
 
@@ -316,13 +317,14 @@ class CroniterTest(base.TestCase):
         wildcard = ['*']
         m, h, d, mon, dow, s = range(6)
         # Test each field individually
-        self.assertEqual(croniter('0-59 0 0 0 0').expanded[m], wildcard)
-        self.assertEqual(croniter('0 0-23 0 0 0').expanded[h], wildcard)
-        self.assertEqual(croniter('0 0 0-31 0 0').expanded[d], wildcard)
+        self.assertEqual(croniter('0-59 0 0 1 0').expanded[m], wildcard)
+        self.assertEqual(croniter('0 0-23 0 1 0').expanded[h], wildcard)
+        self.assertEqual(croniter('0 0 0-31 1 0').expanded[d], wildcard)
         self.assertEqual(croniter('0 0 0 1-12 0').expanded[mon], wildcard)
-        self.assertEqual(croniter('0 0 0 0 0-6').expanded[dow], wildcard)
-        self.assertEqual(croniter('0 0 0 0 1-7').expanded[dow], wildcard)
-        self.assertEqual(croniter('0 0 0 0 0 0-59').expanded[s], wildcard)
+        self.assertEqual(croniter('0 0 0 1 0-6').expanded[dow], wildcard)
+        self.assertEqual(croniter('0 0 0 1 1-7').expanded[dow], wildcard)
+        self.assertEqual(croniter('0 0 0 1 1-7,sat#3').expanded[dow], wildcard)
+        self.assertEqual(croniter('0 0 0 1 0 0-59').expanded[s], wildcard)
         # Real life examples
         self.assertEqual(croniter('30 1-12,0,10-23 15-21 * fri').expanded[h], wildcard)
         self.assertEqual(croniter('30 1-23,0 15-21 * fri').expanded[h], wildcard)
@@ -991,12 +993,12 @@ class CroniterTest(base.TestCase):
              '2019-01-17 00:00:01',
              '2019-01-18 00:00:02',
              '2019-01-19 00:00:03',
-             '2019-01-20 00:00:04',
              '2019-01-23 00:00:00',
              '2019-01-24 00:00:01',
              '2019-01-25 00:00:02',
              '2019-01-26 00:00:03',
-             '2019-01-27 00:00:04'])
+             '2019-01-30 00:00:00',
+             '2019-01-31 00:00:01'])
 
     def test_mixdow(self):
         base = datetime(2018, 10, 1, 0, 0)
@@ -1102,97 +1104,243 @@ class CroniterTest(base.TestCase):
             '2020-03-29T03:01:00+02:00'])
 
 
-class CroniterRangeTest(base.TestCase):
+    def test_nth_wday_simple(self):
+        f = lambda y,m,w: croniter._get_nth_weekday_of_month(y,m,w)
+        sun, mon, tue, wed, thu, fri, sat = range(7)
 
-    def test_1day_step(self):
-        start = datetime(2016, 12, 2)
-        stop = datetime(2016, 12, 10)
-        fwd = list(croniter_range(start, stop, '0 0 * * *'))
-        self.assertEqual(len(fwd), 9)
-        self.assertEqual(fwd[0], start)
-        self.assertEqual(fwd[-1], stop)
-        # Test the same, but in reverse
-        rev = list(croniter_range(stop, start, '0 0 * * *'))
-        self.assertEqual(len(rev), 9)
-        # Ensure forward/reverse are a mirror image
-        rev.reverse()
-        self.assertEqual(fwd, rev)
+        self.assertEqual(f(2000, 1, mon), (3, 10, 17, 24, 31))
+        self.assertEqual(f(2000, 2, tue), (1, 8, 15, 22, 29)) # Leap year
+        self.assertEqual(f(2000, 3, wed), (1, 8, 15, 22, 29))
+        self.assertEqual(f(2000, 4, thu), (6, 13, 20, 27))
+        self.assertEqual(f(2000, 2, fri), (4, 11, 18, 25))
+        self.assertEqual(f(2000, 2, sat), (5, 12, 19, 26))
 
-    def test_1day_step_no_ends(self):
-        # Test without ends (exclusive)
-        start = datetime(2016, 12, 2)
-        stop = datetime(2016, 12, 10)
-        fwd = list(croniter_range(start, stop, '0 0 * * *', exclude_ends=True))
-        self.assertEqual(len(fwd), 7)
-        self.assertNotEqual(fwd[0], start)
-        self.assertNotEqual(fwd[-1], stop)
-        # Test the same, but in reverse
-        rev = list(croniter_range(stop, start, '0 0 * * *', exclude_ends=True))
-        self.assertEqual(len(rev), 7)
-        self.assertNotEqual(fwd[0], stop)
-        self.assertNotEqual(fwd[-1], start)
+    def test_nth_as_last_wday_simple(self):
+        f = lambda y,m,w: croniter._get_nth_weekday_of_month(y,m,w)[-1]
+        sun, mon, tue, wed, thu, fri, sat = range(7)
+        self.assertEqual(f(2000, 2, tue), 29)
+        self.assertEqual(f(2000, 2, sun), 27)
+        self.assertEqual(f(2000, 2, mon), 28)
+        self.assertEqual(f(2000, 2, wed), 23)
+        self.assertEqual(f(2000, 2, thu), 24)
+        self.assertEqual(f(2000, 2, fri), 25)
+        self.assertEqual(f(2000, 2, sat), 26)
 
-    def test_1month_step(self):
-        start = datetime(1982, 1, 1)
-        stop = datetime(1983, 12, 31)
-        res = list(croniter_range(start, stop, '0 0 1 * *'))
-        self.assertEqual(len(res), 24)
-        self.assertEqual(res[0], start)
-        self.assertEqual(res[5].day, 1)
-        self.assertEqual(res[-1], datetime(1983, 12, 1))
+    def test_wdom_core_leap_year(self):
+        f = lambda y,m,w: croniter._get_nth_weekday_of_month(y,m,w)[-1]
+        sun, mon, tue, wed, thu, fri, sat = range(7)
+        self.assertEqual(f(2000, 2, tue), 29)
+        self.assertEqual(f(2000, 2, sun), 27)
+        self.assertEqual(f(2000, 2, mon), 28)
+        self.assertEqual(f(2000, 2, wed), 23)
+        self.assertEqual(f(2000, 2, thu), 24)
+        self.assertEqual(f(2000, 2, fri), 25)
+        self.assertEqual(f(2000, 2, sat), 26)
 
-    def test_1minute_step_float(self):
-        start = datetime(2000, 1, 1, 0, 0)
-        stop =  datetime(2000, 1, 1, 0, 1)
-        res = list(croniter_range(start, stop, '* * * * *', ret_type=float))
-        self.assertEqual(len(res), 2)
-        self.assertEqual(res[0], 946684800.0)
-        self.assertEqual(res[-1] - res[0], 60)
+    def test_lwom_friday(self):
+        it = croniter("0 0 * * L5", datetime(1987, 1, 15), ret_type=datetime)
+        items = [next(it) for i in range(12)]
+        self.assertListEqual(items, [
+            datetime(1987, 1, 30),
+            datetime(1987, 2, 27),
+            datetime(1987, 3, 27),
+            datetime(1987, 4, 24),
+            datetime(1987, 5, 29),
+            datetime(1987, 6, 26),
+            datetime(1987, 7, 31),
+            datetime(1987, 8, 28),
+            datetime(1987, 9, 25),
+            datetime(1987, 10, 30),
+            datetime(1987, 11, 27),
+            datetime(1987, 12, 25),
+        ])
 
-    def test_auto_ret_type(self):
-        data = [
-            (datetime(2019, 1, 1), datetime(2020, 1, 1), datetime),
-            (1552252218.0, 1591823311.0, float),
+    def test_lwom_friday_2hours(self):
+        # This works with +/- 'days=1' in proc_day_of_week_last() and I don't know WHY?!?
+        it = croniter("0 1,5 * * L5", datetime(1987, 1, 15), ret_type=datetime)
+        items = [next(it) for i in range(12)]
+        self.assertListEqual(items, [
+            datetime(1987, 1, 30, 1),
+            datetime(1987, 1, 30, 5),
+            datetime(1987, 2, 27, 1),
+            datetime(1987, 2, 27, 5),
+            datetime(1987, 3, 27, 1),
+            datetime(1987, 3, 27, 5),
+            datetime(1987, 4, 24, 1),
+            datetime(1987, 4, 24, 5),
+            datetime(1987, 5, 29, 1),
+            datetime(1987, 5, 29, 5),
+            datetime(1987, 6, 26, 1),
+            datetime(1987, 6, 26, 5),
+        ])
+
+    def test_lwom_friday_2xh_2xm(self):
+        it = croniter("0,30 1,5 * * L5", datetime(1987, 1, 15), ret_type=datetime)
+        items = [next(it) for i in range(12)]
+        self.assertListEqual(items, [
+            datetime(1987, 1, 30, 1, 0),
+            datetime(1987, 1, 30, 1, 30),
+            datetime(1987, 1, 30, 5, 0),
+            datetime(1987, 1, 30, 5, 30),
+            datetime(1987, 2, 27, 1, 0),
+            datetime(1987, 2, 27, 1, 30),
+            datetime(1987, 2, 27, 5, 0),
+            datetime(1987, 2, 27, 5, 30),
+            datetime(1987, 3, 27, 1, 0),
+            datetime(1987, 3, 27, 1, 30),
+            datetime(1987, 3, 27, 5, 0),
+            datetime(1987, 3, 27, 5, 30),
+        ])
+
+    def test_lwom_saturday_rev(self):
+        it = croniter("0 0 * * L6", datetime(2017, 12, 31), ret_type=datetime, is_prev=True)
+        items = [next(it) for i in range(12)]
+        self.assertListEqual(items, [
+            datetime(2017, 12, 30),
+            datetime(2017, 11, 25),
+            datetime(2017, 10, 28),
+            datetime(2017, 9, 30),
+            datetime(2017, 8, 26),
+            datetime(2017, 7, 29),
+            datetime(2017, 6, 24),
+            datetime(2017, 5, 27),
+            datetime(2017, 4, 29),
+            datetime(2017, 3, 25),
+            datetime(2017, 2, 25),
+            datetime(2017, 1, 28),
+        ])
+
+    def test_lwom_tue_thu(self):
+        it = croniter("0 0 * * L2,L4", datetime(2016, 6, 1), ret_type=datetime)
+        items = [next(it) for i in range(10)]
+        self.assertListEqual(items, [
+            datetime(2016, 6, 28),
+            datetime(2016, 6, 30),
+            datetime(2016, 7, 26),
+            datetime(2016, 7, 28),
+            datetime(2016, 8, 25),  # last tuesday comes before the last thursday
+            datetime(2016, 8, 30),
+            datetime(2016, 9, 27),
+            datetime(2016, 9, 29),
+            datetime(2016, 10, 25),
+            datetime(2016, 10, 27),
+        ])
+
+    def test_hash_mixup_all_fri_3rd_sat(self):
+        # It appears that it's not possible to MIX a literal dow with a `dow#n` format
+        cron_a = "0 0 * * 6#3"
+        cron_b = "0 0 * * 5"
+        cron_c = "0 0 * * 5,6#3"
+        start = datetime(2021, 3, 1)
+        expect_a = [ datetime(2021, 3, 20) ]
+        expect_b = [
+            datetime(2021, 3, 5),
+            datetime(2021, 3, 12),
+            datetime(2021, 3, 19),
+            datetime(2021, 3, 26),
         ]
-        for start, stop, rtype in data:
-            ret = list(croniter_range(start, stop, "0 0 * * *"))
-            self.assertIsInstance(ret[0], rtype)
+        expect_c = sorted(set(expect_a) & set(expect_b))
+        def getn(expr, n):
+            it = croniter(expr, start, ret_type=datetime)
+            return [next(it) for i in range(n)]
+        self.assertListEqual(getn(cron_a, 1), expect_a)
+        self.assertListEqual(getn(cron_b, 4), expect_b)
+        with self.assertRaises(CroniterUnsupportedSyntaxError):
+            self.assertListEqual(getn(cron_c, 5), expect_c)
 
-    def test_input_type_exceptions(self):
-        dt_start1 = datetime(2019, 1, 1)
-        dt_stop1 = datetime(2020, 1, 1)
-        f_start1 = 1552252218.0
-        f_stop1 = 1591823311.0
-        # Mix start/stop types
-        with self.assertRaises(TypeError):
-            list(croniter_range(dt_start1, f_stop1, "0 * * * *"), ret_type=datetime)
-        with self.assertRaises(TypeError):
-            list(croniter_range(f_start1, dt_stop1, "0 * * * *"))
+    def test_lwom_mixup_all_fri_last_sat(self):
+        # Based on the failure of test_hash_mixup_all_fri_3rd_sat, we should expect this to fail too as this implementation simply extends nth_weekday_of_month
+        cron_a = "0 0 * * L6"
+        cron_b = "0 0 * * 5"
+        cron_c = "0 0 * * 5,L6"
+        start = datetime(2021, 3, 1)
+        expect_a = [ datetime(2021, 3, 27) ]
+        expect_b = [
+            datetime(2021, 3, 5),
+            datetime(2021, 3, 12),
+            datetime(2021, 3, 19),
+            datetime(2021, 3, 26),
+        ]
+        expect_c = sorted(set(expect_a) | set(expect_b))
+        def getn(expr, n):
+            it = croniter(expr, start, ret_type=datetime)
+            return [next(it) for i in range(n)]
+        self.assertListEqual(getn(cron_a, 1), expect_a)
+        self.assertListEqual(getn(cron_b, 4), expect_b)
+        with self.assertRaises(CroniterUnsupportedSyntaxError):
+            self.assertListEqual(getn(cron_c, 5), expect_c)
 
-    def test_timezone_dst(self):
-        """ Test across DST transition, which technially is a timzone change. """
-        tz = pytz.timezone("US/Eastern")
-        start = tz.localize(datetime(2020, 10, 30))
-        stop =  tz.localize(datetime(2020, 11, 10))
-        res = list(croniter_range(start, stop, '0 0 * * *'))
-        self.assertNotEqual(res[0].tzinfo, res[-1].tzinfo)
-        self.assertEqual(len(res), 12)
+    def test_lwom_mixup_firstlast_sat(self):
+        # First saturday, last saturday
+        start = datetime(2021, 3, 1)
+        cron_a = "0 0 * * 6#1"
+        cron_b = "0 0 * * L6"
+        cron_c = "0 0 * * L6,6#1"
+        expect_a = [
+            datetime(2021, 3, 6),
+            datetime(2021, 4, 3),
+            datetime(2021, 5, 1),
+        ]
+        expect_b = [
+            datetime(2021, 3, 27),
+            datetime(2021, 4, 24),
+            datetime(2021, 5, 29),
+        ]
+        expect_c = sorted(expect_a + expect_b)
+        def getn(expr, n):
+            it = croniter(expr, start, ret_type=datetime)
+            return [next(it) for i in range(n)]
+        self.assertListEqual(getn(cron_a, 3), expect_a)
+        self.assertListEqual(getn(cron_b, 3), expect_b)
+        self.assertListEqual(getn(cron_c, 6), expect_c)
 
-    def test_extra_hour_day_prio(self):
-        def datetime_tz(*args, **kw):
-            """ Defined this in another branch.  single-use-version """
-            tzinfo = kw.pop("tzinfo")
-            return tzinfo.localize(datetime(*args))
-        tz = pytz.timezone("US/Eastern")
-        cron = "0 3 * * *"
-        start = datetime_tz(2020, 3, 7, tzinfo=tz)
-        end = datetime_tz(2020, 3, 11, tzinfo=tz)
-        ret = [ i.isoformat() for i in croniter_range(start, end, cron) ]
-        self.assertEqual(ret, [
-            "2020-03-07T03:00:00-05:00",
-            "2020-03-08T03:00:00-04:00",
-            "2020-03-09T03:00:00-04:00",
-            "2020-03-10T03:00:00-04:00"])
+    def test_lwom_mixup_4th_and_last(self):
+        # 4th and last monday
+        start = datetime(2021, 11, 1)
+        cron_a = "0 0 * * 1#4"
+        cron_b = "0 0 * * L1"
+        cron_c = "0 0 * * 1#4,L1"
+        expect_a = [
+            datetime(2021, 11, 22),
+            datetime(2021, 12, 27),
+            datetime(2022, 1, 24)
+        ]
+        expect_b = [
+            datetime(2021, 11, 29),
+            datetime(2021, 12, 27),
+            datetime(2022, 1, 31),
+        ]
+        expect_c = sorted(set(expect_a) | set(expect_b))
+
+        def getn(expr, n):
+            it = croniter(expr, start, ret_type=datetime)
+            return [next(it) for i in range(n)]
+
+        self.assertListEqual(getn(cron_a, 3), expect_a)
+        self.assertListEqual(getn(cron_b, 3), expect_b)
+        self.assertListEqual(getn(cron_c, 5), expect_c)
+
+    def test_nth_out_of_range(self):
+        with self.assertRaises(CroniterBadCronError):
+            croniter("0 0 * * 1#7")
+        with self.assertRaises(CroniterBadCronError):
+            croniter("0 0 * * 1#0")
+
+    def test_last_out_of_range(self):
+        with self.assertRaises(CroniterBadCronError):
+            croniter("0 0 * * L-1")
+        with self.assertRaises(CroniterBadCronError):
+            croniter("0 0 * * L8")
+
+    def test_issue_47(self):
+        base = datetime(2021, 3, 30, 4, 0)
+        itr = croniter('0 6 30 3 *', base)
+        prev1 = itr.get_prev(datetime)
+        self.assertEqual(prev1.year, base.year-1)
+        self.assertEqual(prev1.month, 3)
+        self.assertEqual(prev1.day, 30)
+        self.assertEqual(prev1.hour, 6)
+        self.assertEqual(prev1.minute, 0)
 
     def test_issue_142_dow(self):
         ret = []
@@ -1246,23 +1394,11 @@ class CroniterRangeTest(base.TestCase):
         it = croniter(cron, start, day_or=False, max_years_between_matches=5)
         self.assertEqual(it.get_next(datetime), datetime(2025, 1, 8, 13))
 
-    def test_issue145_range(self):
-        cron = "0 13 8 1,4,7,10 wed"
-        matches = list(croniter_range(datetime(2020, 1, 1), datetime(2020, 12, 31), cron, day_or=False))
-        self.assertEqual(len(matches), 3)
-        self.assertEqual(matches[0], datetime(2020, 1, 8, 13))
-        self.assertEqual(matches[1], datetime(2020, 4, 8, 13))
-        self.assertEqual(matches[2], datetime(2020, 7, 8, 13))
-
-        # No matches within this range; therefore expect empty list
-        matches = list(croniter_range(datetime(2020, 9, 30), datetime(2020, 10, 30), cron, day_or=False))
-        self.assertEqual(len(matches), 0)
-
     def test_explicit_year_forward(self):
         start = datetime(2020, 9, 24)
         cron = "0 13 8 1,4,7,10 wed"
 
-        # Expect exception because no explict range was provided.  Therefore, the caller should be made aware that an implicit limit was hit.
+        # Expect exception because no explicit range was provided.  Therefore, the caller should be made aware that an implicit limit was hit.
         ccron = croniter(cron, start, day_or=False)
         ccron._max_years_between_matches = 1
         iterable = ccron.all_next()
@@ -1273,29 +1409,11 @@ class CroniterRangeTest(base.TestCase):
         n = next(iterable)
         self.assertEqual(n, datetime(2025, 1, 8, 13))
 
-        # If the explictly given lookahead isn't enough to reach the next date, that's fine.  The caller specified the maximum gap, so no just stop iteration
+        # If the explicitly given lookahead isn't enough to reach the next date, that's fine.  The caller specified the maximum gap, so no just stop iteration
         iterable = croniter(cron, start, day_or=False, max_years_between_matches=2).all_next(datetime)
         with self.assertRaises(StopIteration):
             next(iterable)
 
-    def test_explicit_year_reverse(self):
-        start = datetime(2025, 1, 1)
-        cron = "0 13 8 1,4,7,10 wed"
-
-        ccron = croniter(cron, start, day_or=False)
-        ccron._max_years_between_matches = 1
-        iterable = ccron.all_prev()
-        with self.assertRaises(CroniterBadDateError):
-            next(iterable)
-
-        iterable = croniter(cron, start, day_or=False, max_years_between_matches=5).all_prev(datetime)
-        n = next(iterable)
-        self.assertEqual(n, datetime(2020, 7, 8, 13))
-
-        iterable = croniter(cron, start, day_or=False, max_years_between_matches=2).all_prev()
-        with self.assertRaises(StopIteration):
-            next(iterable)
-
     def test_issue151(self):
         """."""
         self.assertTrue(
@@ -1315,11 +1433,14 @@ class CroniterRangeTest(base.TestCase):
 
     def test_confirm_sort(self):
         m, h, d, mon, dow, s = range(6)
-        self.assertListEqual(croniter('0 8,22,10,23 0 0 0').expanded[h], [8, 10, 22, 23])
-        self.assertListEqual(croniter('0 0 25-L 0 0').expanded[d], [25, 26, 27, 28, 29, 30, 31])
+        self.assertListEqual(croniter('0 8,22,10,23 0 1 0').expanded[h], [8, 10, 22, 23])
+        self.assertListEqual(croniter('0 0 25-L 1 0').expanded[d], [25, 26, 27, 28, 29, 30, 31])
         self.assertListEqual(croniter("1 1 7,14,21,L * *").expanded[d], [7, 14, 21, "l"])
         self.assertListEqual(croniter("0 0 * * *,sat#3").expanded[dow], ["*", 6])
 
+    def test_issue_k6(self):
+        self.assertRaises(CroniterBadCronError, croniter, '0 0 0 0 0')
+
 
 if __name__ == '__main__':
     unittest.main()
